@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2018 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 
@@ -37,18 +37,41 @@
 #define MAXPATHLEN 1024
 
 /* definitions for fontsource and font number table entries */
+#define pjl_fontsource_body\
+    char designator[2]; \
+    PJCONST char *pathname; \
+    PJCONST char *fontnumber
+
+#define PJCONST
 typedef struct pjl_fontsource
 {
-    char designator[2];
-    char *pathname;
-    char *fontnumber;
+  pjl_fontsource_body;
 } pjl_fontsource_t;
+#undef PJCONST
+
+#define PJCONST const
+typedef struct pjl_fontsource_default
+{
+  pjl_fontsource_body;
+} pjl_fontsource_default_t;
+#undef PJCONST
 
 /* definitions for variable names and values */
+#define pjl_envir_var_body\
+  PJCONST char *var; \
+  PJCONST char *value
+
+#define PJCONST
 typedef struct pjl_envir_var_s {
-  char *var;
-  char *value;
+  pjl_envir_var_body;
 } pjl_envir_var_t;
+#undef PJCONST
+
+#define PJCONST const
+typedef struct pjl_envir_var_default_s {
+  pjl_envir_var_body;
+} pjl_envir_var_default_t;
+#undef PJCONST
 
 /* the pjl current environment and the default user environment.  Note
    the default environment should be stored in NVRAM on embedded print
@@ -75,7 +98,7 @@ typedef struct pjl_parser_state_s
 /* provide factory defaults for pjl commands.  Note these are not pjl
    defaults but initial values set in the printer when shipped.  In an
    embedded system these would be defined in ROM */
-static pjl_envir_var_t pjl_factory_defaults[] = {
+static pjl_envir_var_default_t pjl_factory_defaults[] = {
     {"formlines", "60"},
     {"formlines_set", "off"},
     {"widea4", "no"},
@@ -122,9 +145,9 @@ static pjl_envir_var_t pjl_factory_defaults[] = {
    host based systems.  Entries are seperated with a semi-colon.  Note
    there is some unnecessary overlap in the factory default and font
    source table. */
-static pjl_fontsource_t pjl_fontsource_table[] = {
+static pjl_fontsource_default_t pjl_fontsource_table[] = {
     {"I",
-     "%rom%ttfonts/;fonts/;urwfonts/;/windows/fonts/;/win95/fonts/;/winnt/fonts/", ""},
+     "%rom%ttfonts/;urwfonts/;pcl/urwfonts/;ghostpdl/pcl/urwfonts/;/windows/fonts/;", ""},
     {"C", "CART0/", ""},
     {"C1", "CART1/", ""},
     {"C2", "CART2/", ""},
@@ -640,7 +663,11 @@ pjl_fsinit(pjl_parser_state_t * pst, char *pathname)
     pjl_parsed_filename_to_string(fname, pathname);
     if (pjl_verify_file_operation(pst, fname) < 0)
         return -1;
+#ifdef GS_NO_FILESYSTEM
+    return -1;
+#else
     return mkdir(fname, 0777);
+#endif
 }
 
 /* make a pjl directory */
@@ -652,7 +679,11 @@ pjl_fsmkdir(pjl_parser_state_t * pst, char *pathname)
     pjl_parsed_filename_to_string(fname, pathname);
     if (pjl_verify_file_operation(pst, fname) < 0)
         return -1;
+#ifdef GS_NO_FILESYSTEM
+    return -1;
+#else
     return mkdir(fname, 0777);
+#endif
 }
 
 /* query a file in the pjl sandbox */
@@ -806,8 +837,10 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
     token = (char *)gs_alloc_bytes(pst->mem, bufsize, "working buffer for PJL parsing");
     if (token == 0)
         return -1;
+    
     memset(token, 0x00, bufsize);
-
+    memset(pathname, 0x00, sizeof(pathname));
+           
     /* reset the line position to the beginning of the line */
     pst->pos = 0;
     /* all pjl commands start with the pjl prefix @PJL */
@@ -894,7 +927,8 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
                     if ((tok = pjl_get_token(pst, token)) == FORMATBINARY) {
                         ;
                     }
-                    if (pjl_get_setting(pst, NAME, token) < 0)
+                    if (pjl_get_setting(pst, NAME, token) < 0 ||
+                        strlen(token) > MAXPATHLEN - 1)
                         return -1;
                     strcpy(pathname, token);
                     if (pjl_get_setting(pst, SIZE, token) < 0)
@@ -909,7 +943,8 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
                     if ((tok = pjl_get_token(pst, token)) == FORMATBINARY) {
                         ;
                     }
-                    if (pjl_get_setting(pst, NAME, token) < 0)
+                    if (pjl_get_setting(pst, NAME, token) < 0 ||
+                        strlen(token) > MAXPATHLEN - 1)
                         return -1;
                     strcpy(pathname, token);
                     if (pjl_get_setting(pst, SIZE, token) < 0)
@@ -919,7 +954,8 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
                     return pjl_set_append_state(pst, pathname, size);
                 }
             case FSDELETE:
-                if (pjl_get_setting(pst, NAME, token) < 0)
+                if (pjl_get_setting(pst, NAME, token) < 0 ||
+                    strlen(token) > MAXPATHLEN - 1)
                     return -1;
                 strcpy(pathname, token);
                 gs_free_object(pst->mem, token, "working buffer for PJL parsing");
@@ -929,7 +965,8 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
 
                     int count;
 
-                    if (pjl_get_setting(pst, NAME, token) < 0) {
+                    if (pjl_get_setting(pst, NAME, token) < 0 ||
+                        strlen(token) > MAXPATHLEN - 1) {
                         gs_free_object(pst->mem, token, "working buffer for PJL parsing");
                         return -1;
                     }
@@ -948,7 +985,8 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
                     return pjl_fsdirlist(pst, pathname, entry, count);
                 }
             case FSINIT:
-                if (pjl_get_setting(pst, VOLUME, token) < 0) {
+                if (pjl_get_setting(pst, VOLUME, token) < 0 ||
+                    strlen(token) > MAXPATHLEN - 1) {
                     gs_free_object(pst->mem, token, "working buffer for PJL parsing");
                     return -1;
                 }
@@ -956,7 +994,8 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
                 gs_free_object(pst->mem, token, "working buffer for PJL parsing");
                 return pjl_fsinit(pst, pathname);
             case FSMKDIR:
-                if (pjl_get_setting(pst, NAME, token) < 0) {
+                if (pjl_get_setting(pst, NAME, token) < 0 ||
+                    strlen(token) > MAXPATHLEN - 1) {
                     gs_free_object(pst->mem, token, "working buffer for PJL parsing");
                     return -1;
                 }
@@ -964,7 +1003,8 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
                 gs_free_object(pst->mem, token, "working buffer for PJL parsing");
                 return pjl_fsmkdir(pst, pathname);
             case FSQUERY:
-                if (pjl_get_setting(pst, NAME, token) < 0) {
+                if (pjl_get_setting(pst, NAME, token) < 0 ||
+                    strlen(token) > MAXPATHLEN - 1) {
                     gs_free_object(pst->mem, token, "working buffer for PJL parsing");
                     return -1;
                 }
@@ -979,7 +1019,8 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
                     if ((tok = pjl_get_token(pst, token)) == FORMATBINARY) {
                         ;
                     }
-                    if (pjl_get_setting(pst, NAME, token) < 0) {
+                    if (pjl_get_setting(pst, NAME, token) < 0 ||
+                        strlen(token) > MAXPATHLEN - 1) {
                         gs_free_object(pst->mem, token, "working buffer for PJL parsing");
                         return -1;
                     }
@@ -995,7 +1036,7 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
                     }
                     size = pjl_vartoi(token);
                     gs_free_object(pst->mem, token, "working buffer for PJL parsing");
-                    return pjl_fsupload(pst, pathname, size, offset);
+                    return pjl_fsupload(pst, pathname, offset, size);
                 }
             default:
                 gs_free_object(pst->mem, token, "working buffer for PJL parsing");
@@ -1003,7 +1044,7 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
         }
     }
     gs_free_object(pst->mem, token, "working buffer for PJL parsing");
-    return (tok == DONE ? 0 : -1);
+    return 0;
 }
 
 /* get a file from 0: or 1: volume */
@@ -1047,6 +1088,7 @@ int
 pjl_get_named_resource(pjl_parser_state * pst, char *name, byte * data)
 {
     long int size;
+    int code = 0;
     FILE *fp = get_fp(pst, name);
 
     if (fp == NULL)
@@ -1054,11 +1096,11 @@ pjl_get_named_resource(pjl_parser_state * pst, char *name, byte * data)
     fseek(fp, 0L, SEEK_END);
     size = ftell(fp);
     rewind(fp);
-    if (size != fread(data, 1, size, fp)) {
-        fclose(fp);
-        return -1;
+    if (size < 0 || (size != fread(data, 1, size, fp))) {
+        code = -1;
     }
-    return 0;
+    fclose(fp);
+    return code;
 }
 
 /* set the initial environment to the default environment, this should
@@ -1224,13 +1266,18 @@ pjl_process(pjl_parser_state * pst, void *pstate, stream_cursor_read * pr)
             ++p;
             /* null terminate, parse and set the pjl state */
             pst->line[pst->pos] = '\0';
-            pjl_parse_and_process_line(pst);
+            /*
+             * NB PJL continues not considering normal errors but we
+             * ignore memory failure errors here and shouldn't.
+             */
+            (void)pjl_parse_and_process_line(pst);
             pst->pos = 0;
             continue;
         }
         
         if (!legal_pjl_char(p[1])) {
             code = 1;
+            ++p;
             break;
         }
 
@@ -1238,14 +1285,11 @@ pjl_process(pjl_parser_state * pst, void *pstate, stream_cursor_read * pr)
         /* Always leave room for a terminator. */
         if (pst->pos == pst->line_size) {
             char *temp = (char *)gs_alloc_bytes (pst->mem, pst->line_size + 256, "pjl_state increase line buffer");
-
-            if (p) {
-                memcpy(temp, pst->line, pst->line_size);
-                gs_free_object(pst->mem, pst->line, "pjl_state line buffer");
-                pst->line = temp;
-                pst->line_size += 256;
-                pst->line[pst->pos] = p[1], pst->pos++;
-            }
+            memcpy(temp, pst->line, pst->line_size);
+            gs_free_object(pst->mem, pst->line, "pjl_state line buffer");
+            pst->line = temp;
+            pst->line_size += 256;
+            pst->line[pst->pos] = p[1], pst->pos++;
         } else
             pst->line[pst->pos] = p[1], pst->pos++;
         ++p;
@@ -1282,63 +1326,50 @@ pjl_skip_to_uel(stream_cursor_read * pr)
 static const struct
 {
     const char *symname;
-    const char *pcl_selectcode;
+    const int  pcl_code;
 } symbol_sets[] = {
-    {"ROMAN8", "8U"},
-    {"ISOL1", "0N"},
-    {"ISOL2", "2N"},
-    {"ISOL5", "5N"},
-    {"PC8", "10U"},
-    {"PC8DN", "11U"},
-    {"PC850", "12U"},
-    {"PC852", "17U"},
-    {"PC8TK", "9T"},             /* pc-8 turkish ?? not sure */
-    {"WINL1", "9U"},
-    {"WINL2", "9E"},
-    {"WINL5", "5T"},
-    {"DESKTOP", "7J"},
-    {"PSTEXT", "10J"},
-    {"VNINTL", "13J"},
-    {"VNUS", "14J"},
-    {"MSPUBL", "6J"},
-    {"MATH8", "8M"},
-    {"PSMATH", "5M"},
-    {"VNMATH", "6M"},
-    {"PIFONT", "15U"},
-    {"LEGAL", "1U"},
-    {"ISO4", "1E"},
-    {"ISO6", "0U"},
-    {"ISO11", "0S"},
-    {"ISO15", "0I"},
-    {"ISO17", "2S"},
-    {"ISO21", "1G"},
-    {"ISO60", "OD"},
-    {"ISO69", "1F"},
-    {"WIN30", NULL},             /* don't know */
-    {"WIN31J", NULL},            /* don't know */
-    {"GB2312", NULL},            /* don't know */
-    {NULL, NULL}
+    {"ROMAN8",  277},
+    {"ISOL1",    14},
+    {"ISOL2",    78},
+    {"ISOL5",   174},
+    {"PC8",     341},
+    {"PC8DN",   373},
+    {"PC850",   405},
+    {"PC852",   565},
+    {"PC8TK",   308},
+    {"WINL1",   309},
+    {"WINL2",   293},
+    {"WINL5",   180},
+    {"DESKTOP", 234},
+    {"PSTEXT",  330},
+    {"VNINTL",  426},
+    {"VNUS",    458},
+    {"MSPUBL",  202},
+    {"MATH8",   269},
+    {"PSMATH",  173},
+    {"VNMATH",  205},
+    {"PIFONT",  501},
+    {"LEGAL",    53},
+    {"ISO4",     37},
+    {"ISO6",     21},
+    {"ISO11",    19},
+    {"ISO15",     9},
+    {"ISO17",    83},
+    {"ISO21",    39},
+    {"ISO60",     4},
+    {"ISO69",    38},
+    {NULL,       -1}
 };
 
-/* map a pjl symbol table name to a pcl symbol table name */
+/* map a pjl symbol table name to a pcl symbol table code */
 int
 pjl_map_pjl_sym_to_pcl_sym(const char *symname)
 {
     int i;
 
     for (i = 0; symbol_sets[i].symname; i++)
-        if (!pjl_compare(symname, symbol_sets[i].symname)) {
-            /* convert the character code to it's integer
-               representation.  NB peculiar code! */
-            char pcl_symbol[4], chr;
-            int char_pos;
-
-            strcpy(pcl_symbol, symbol_sets[i].pcl_selectcode);
-            char_pos = strlen(pcl_symbol) - 1;
-            chr = pcl_symbol[char_pos];
-            pcl_symbol[char_pos] = '\0';
-            return (atoi(pcl_symbol) << 5) + chr - 64;
-        }
+        if (!pjl_compare(symname, symbol_sets[i].symname))
+            return symbol_sets[i].pcl_code;
     return -1;
 }
 
@@ -1384,26 +1415,27 @@ static int set_pjl_defaults(gs_memory_t * mem, pjl_envir_var_t **def, pjl_envir_
     pjl_envir_var_t *pjl_def;
     char *key, *value, *newkey, *newvalue;
     int i = 0;
+    int limit = 0;
 
-    while (from[i].var && from[i].var[0] != 0x00)
-        i++;
+    while (from[limit].var && from[limit].var[0] != 0x00)
+        limit++;
 
     pjl_def = (pjl_envir_var_t *) gs_alloc_bytes(mem,
                                                   sizeof
-                                                  (pjl_envir_var_t) * (i + 1),
+                                                  (pjl_envir_var_t) * (limit + 1),
                                                   "pjl_envir");
     if (!pjl_def)
         return -1;
 
-    memset(pjl_def, 0x00, sizeof(pjl_envir_var_t) * (i + 1));
-    i--;
-    while (i >= 0) {
+    memset(pjl_def, 0x00, sizeof(pjl_envir_var_t) * (limit + 1));
+    while (i < limit) {
         key = from[i].var;
         value = from[i].value;
 
         newkey = (char *)gs_alloc_bytes(mem, strlen(key) + 1, "new_pjl_defaults, key");
         newvalue = (char *)gs_alloc_bytes(mem, strlen(value) + 1, "new_pjl_defaults, value");
         if (!newkey || !newvalue) {
+            gs_free_object(mem, newkey, "new_pjl_defaults, key");
             free_pjl_defaults(mem, &pjl_def);
             return -1;
         }
@@ -1412,7 +1444,7 @@ static int set_pjl_defaults(gs_memory_t * mem, pjl_envir_var_t **def, pjl_envir_
         pjl_def[i].var = newkey;
         pjl_def[i].value = newvalue;
 
-        i--;
+        i++;
     }
     *def = pjl_def;
     return 0;
@@ -1428,25 +1460,26 @@ static int set_pjl_environment(gs_memory_t * mem, pjl_envir_var_t **env, pjl_env
     pjl_envir_var_t *pjl_env;
     char *key, *value, *newkey, *newvalue;
     int i = 0;
+    int limit = 0;
 
-    while (from[i].var && from[i].var[0] != 0x00)
-        i++;
+    while (from[limit].var && from[limit].var[0] != 0x00)
+        limit++;
 
     pjl_env = (pjl_envir_var_t *) gs_alloc_bytes(mem,
                                                   sizeof
-                                                  (pjl_envir_var_t) * (i + 1),
+                                                  (pjl_envir_var_t) * (limit + 1),
                                                   "pjl_envir");
     if (!pjl_env)
         return -1;
 
-    memset(pjl_env, 0x00, sizeof(pjl_envir_var_t) * (i + 1));
-    i--;
-    while (i >= 0) {
+    memset(pjl_env, 0x00, sizeof(pjl_envir_var_t) * (limit + 1));
+    while (i < limit) {
         key = from[i].var;
         value = from[i].value;
         newkey = (char *)gs_alloc_bytes(mem, strlen(key) + 1, "pjl_envir, key");
         newvalue = (char *)gs_alloc_bytes(mem, strlen(value) + 1, "pjl_envir, value");
         if (!newkey || !newvalue) {
+            gs_free_object(mem, newkey, "pjl_envir, key");
             free_pjl_environment(mem, &pjl_env);
             return -1;
         }
@@ -1455,7 +1488,7 @@ static int set_pjl_environment(gs_memory_t * mem, pjl_envir_var_t **env, pjl_env
         pjl_env[i].var = newkey;
         pjl_env[i].value = newvalue;
 
-        i--;
+        i++;
     }
     *env = pjl_env;
     return 0;
@@ -1471,20 +1504,20 @@ static int set_pjl_fontsource(gs_memory_t * mem, pjl_fontsource_t **fontenv, pjl
     pjl_fontsource_t *pjl_fontenv;
     char *key, *value, *newkey, *newvalue;
     int i = 0;
+    int limit = 0;
 
-    while (from[i].pathname && from[i].pathname[0] != 0x00)
-        i++;
+    while (from[limit].pathname && from[limit].pathname[0] != 0x00)
+        limit++;
 
     pjl_fontenv = (pjl_fontsource_t *) gs_alloc_bytes(mem,
                                                       sizeof
-                                                      (pjl_fontsource_t) * (i + 1),
+                                                      (pjl_fontsource_t) * (limit + 1),
                                                       "font_envir");
     if (!pjl_fontenv)
         return -1;
 
-    memset(pjl_fontenv, 0x00, sizeof(pjl_fontsource_t) * (i + 1));
-    i--;
-    while (i >= 0) {
+    memset(pjl_fontenv, 0x00, sizeof(pjl_fontsource_t) * (limit + 1));
+    while (i < limit) {
         key = from[i].pathname;
         value = from[i].fontnumber;
 
@@ -1508,7 +1541,7 @@ static int set_pjl_fontsource(gs_memory_t * mem, pjl_fontsource_t **fontenv, pjl
         }
         memcpy(pjl_fontenv[i].designator, pjl_fontsource_table[i].designator, 2);
 
-        i--;
+        i++;
     }
     *fontenv = pjl_fontenv;
     return 0;
@@ -1524,20 +1557,20 @@ static int set_pjl_default_fontsource(gs_memory_t * mem, pjl_fontsource_t **font
     pjl_fontsource_t *pjl_fontdef;
     char *key, *value, *newkey, *newvalue;
     int i = 0;
+    int limit = 0;
 
-    while (from[i].pathname && from[i].pathname[0] != 0x00)
-        i++;
+    while (from[limit].pathname && from[limit].pathname[0] != 0x00)
+        limit++;
 
     pjl_fontdef = (pjl_fontsource_t *) gs_alloc_bytes(mem,
                                                       sizeof
-                                                      (pjl_fontsource_t) * (i + 1),
+                                                      (pjl_fontsource_t) * (limit + 1),
                                                       "new_font_defaults");
     if (!pjl_fontdef)
         return -1;
 
-    memset(pjl_fontdef, 0x00, sizeof(pjl_fontsource_t) * (i + 1));
-    i--;
-    while (i >= 0) {
+    memset(pjl_fontdef, 0x00, sizeof(pjl_fontsource_t) * (limit + 1));
+    while (i < limit) {
         key = from[i].pathname;
         value = from[i].fontnumber;
 
@@ -1561,7 +1594,7 @@ static int set_pjl_default_fontsource(gs_memory_t * mem, pjl_fontsource_t **font
         }
         memcpy(pjl_fontdef[i].designator, pjl_fontsource_table[i].designator, 2);
 
-        i--;
+        i++;
     }
     *fontdef = pjl_fontdef;
     return 0;
@@ -1587,6 +1620,9 @@ int free_pjl_environment(gs_memory_t * mem, pjl_envir_var_t **env)
     int i=0;
     pjl_envir_var_t *pjl_env = *env;
 
+    if (pjl_env == NULL)
+        return 0;
+
     while (pjl_env[i].var) {
         gs_free_object(mem, pjl_env[i].var, "free pjl_environment key");
         gs_free_object(mem, pjl_env[i].value, "free pjl_environment value");
@@ -1602,6 +1638,9 @@ int free_pjl_fontsource(gs_memory_t * mem, pjl_fontsource_t **fontenv)
     int i=0;
     pjl_fontsource_t *pjl_fontenv = *fontenv;
 
+    if (pjl_fontenv == NULL)
+        return 0;
+
     while (pjl_fontenv[i].pathname) {
         gs_free_object(mem, pjl_fontenv[i].pathname, "pjl_font_envir pathname");
         gs_free_object(mem, pjl_fontenv[i].fontnumber, "pjl_font_envir fontnumber");
@@ -1675,26 +1714,16 @@ pjl_process_init(gs_memory_t * mem)
     /* initialize the default and initial pjl environment.  We assume
        that these are the same layout as the factory defaults. */
     if (set_pjl_defaults_to_factory(mem, &pjl_def) < 0) {
-        gs_free_object(mem, pjlstate, "pjl_state");
-        return NULL;
+        goto fail1;
     }
     if (set_pjl_environment_to_factory(mem, &pjl_env) < 0) {
-        free_pjl_defaults(mem, &pjl_def);
-        gs_free_object(mem, pjlstate, "pjl_state");
-        return NULL;
+        goto fail2;
     }
     if (set_pjl_fontsource_to_factory(mem, &pjl_fontenv) < 0) {
-        free_pjl_defaults(mem, &pjl_def);
-        free_pjl_environment(mem, &pjl_env);
-        gs_free_object(mem, pjlstate, "pjl_state");
-        return NULL;
+        goto fail3;
     }
     if (set_pjl_default_fontsource_to_factory(mem, &pjl_fontdef) < 0) {
-        free_pjl_defaults(mem, &pjl_def);
-        free_pjl_environment(mem, &pjl_env);
-        free_pjl_fontsource(mem, &pjl_fontenv);
-        gs_free_object(mem, pjlstate, "pjl_state");
-        return NULL;
+        goto fail4;
     }
 
     /* initialize the font repository data as well */
@@ -1719,6 +1748,17 @@ pjl_process_init(gs_memory_t * mem)
             pjl_permanent_soft_fonts[i] = 0;
     }
     return (pjl_parser_state *) pjlstate;
+
+fail4:
+    free_pjl_fontsource(mem, &pjl_fontenv);
+fail3:
+    free_pjl_environment(mem, &pjl_env);
+fail2:
+    free_pjl_defaults(mem, &pjl_def);
+fail1:
+    gs_free_object(mem, pjlstate->line, "pjl_state line buffer");
+    gs_free_object(mem, pjlstate, "pjl_state");
+    return NULL;
 }
 
 /* case insensitive comparison of two null terminated strings. */
@@ -1733,8 +1773,15 @@ pjl_compare(const pjl_envvar_t * s1, const char *s2)
 
 /* free all memory associated with the PJL state */
 void
-pjl_process_destroy(pjl_parser_state * pst, gs_memory_t * mem)
+pjl_process_destroy(pjl_parser_state * pst)
 {
+    gs_memory_t *mem;
+
+    if (pst == NULL)
+        return;
+
+    mem = pst->mem;
+    
     free_pjl_defaults(mem, &pst->defaults);
     free_pjl_environment(mem, &pst->envir);
     free_pjl_fontsource(mem, &pst->font_envir);
@@ -1742,6 +1789,7 @@ pjl_process_destroy(pjl_parser_state * pst, gs_memory_t * mem)
 
     if (pst->environment_font_path)
         gs_free_object(mem, pst->environment_font_path, "pjl_state");
+    gs_free_object(mem, pst->line, "pjl_state line buffer");
     gs_free_object(mem, pst, "pjl_state");
 }
 
@@ -1800,7 +1848,7 @@ pjl_register_permanent_soft_font_addition(pjl_parser_state * pst)
     /* Find an empty slot in the table.  We have no HP documentation
        that says how a soft font gets associated with a font number */
     int font_num;
-    bool slot_found;
+    bool slot_found = false;
 
     for (font_num = 0; font_num < MAX_PERMANENT_FONTS; font_num++)
         if (!
