@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2018 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 
@@ -116,8 +116,25 @@ typedef struct px_bitmap_enum_s
     stream_DCT_state dct_stream_state;
     jpeg_decompress_data jdd;
     deltarow_state_t deltarow_state;
+
+    /* Entries to deal with 'rebuffering' - the idea of making
+     * input image data appear a different (but equivalent) shape.
+     * Currently just morphing '1xH' bitmaps into 'Hx1' ones. */
+    uint rebuffered_data_per_row; /* equal to data_per_row except when flipping */
+    bool rebuffered;
+    uint rebuffered_width;
+    uint rebuffered_height;
+    int rebuffered_row_pos;
+    bool grayscale;
 } px_bitmap_enum_t;
 
+typedef struct px_begin_image_args_s {
+    uint width, height;
+    int depth;
+    pxeColorMapping_t mapping;
+    real dest_width, dest_height;
+} px_bitmap_args_t;
+
 /* Define our image enumerator. */
 #ifndef px_image_enum_DEFINED
 #  define px_image_enum_DEFINED
@@ -127,31 +144,127 @@ struct px_image_enum_s
 {
     gs_image_t image;
     byte *row;                  /* buffer for one row of data */
-    uint raster;
     void *info;                 /* state structure for driver */
     px_bitmap_enum_t benum;
+    px_bitmap_args_t bi_args;
+    bool enum_started;
 };
 
 gs_private_st_simple(st_px_image_enum, px_image_enum_t, "px_image_enum_t");
 
-/* ---------------- Utilities ---------------- */
+
+static int
+stream_error(stream_state * st, const char *str)
+{
+    return_error(-1);
+}
+
+static int
+px_jpeg_init(px_bitmap_enum_t * benum, px_bitmap_params_t * params, px_args_t * par)
+{
+    stream_cursor_read r;
+    stream_cursor_write w;
+    uint used;
+    const byte *data = par->source.data;
+    stream_DCT_state *ss = (&benum->dct_stream_state);
+    jpeg_decompress_data *jddp = &(benum->jdd);
+    uint avail = par->source.available;
+    int code = 0;
+
+    if (benum->initialized == false) {
+        s_init_state((stream_state *)ss, &s_DCTD_template, benum->mem);
+        ss->report_error = stream_error;
+        s_DCTD_template.set_defaults((stream_state *)ss);
+
+        ss->jpeg_memory = benum->mem;
+        ss->data.decompress = jddp;
+
+        jddp->templat = s_DCTD_template;
+        jddp->memory = benum->mem;
+        jddp->scanline_buffer = NULL;
+        jddp->PassThrough = 0;
+        jddp->PassThroughfn = 0;
+        jddp->device = (void *)0;
+
+        if (gs_jpeg_create_decompress(ss) < 0)
+            return_error(errorInsufficientMemory);
+
+        (*s_DCTD_template.init) ((stream_state *)ss);
+        benum->initialized = true;
+    }
+
+    r.ptr = data - 1;
+    r.limit = r.ptr + avail;
+
+    /* this seems to be the way to just get the header info */
+    w.ptr = w.limit = 0;
+    code =
+        (*s_DCTD_template.process) ((stream_state *)ss, &r, &w, false);
+
+    if (code < 0)
+        goto error;
+
+    used = r.ptr + 1 - data;
+    par->source.data = r.ptr + 1;
+    par->source.available = avail - used;
+    
+    /* Needs more data to complete reading the header */
+    if (ss->phase < 2)
+        return pxNeedData;
+    
+    params->width = jddp->dinfo.output_width;
+    params->height = jddp->dinfo.output_height;
+    {
+        int comps = jddp->dinfo.output_components;
+        if (comps == 1) {
+            params->color_space = eGray;
+            params->depth = 1;
+        } else if (comps == 3) {
+            params->color_space = eRGB;
+            params->depth = 8;
+        }
+        else {
+            code = -1;
+            goto error;
+        }
+        return 0;
+    }
+
+error:
+    gs_jpeg_destroy(ss);
+    return_error(code);
+}
 
 /* Extract the parameters for reading a bitmap image or raster pattern. */
-/* Attributes: pxaColorMapping, pxaColorDepth, pxaSourceWidth, */
-/* pxaSourceHeight, pxaDestinationSize. */
 static int
 begin_bitmap(px_bitmap_params_t * params, px_bitmap_enum_t * benum,
-             const px_args_t * par, const px_state_t * pxs)
+             const px_bitmap_args_t * bar, const px_state_t * pxs, bool is_jpeg, px_args_t * par)
 {
+    int depth;
+    int num_components;
     px_gstate_t *pxgs = pxs->pxgs;
-    int depth = "\001\004\010"[par->pv[1]->value.i];
-    int num_components = (pxgs->color_space == eGray ? 1 : 3);
+    
+    benum->mem = pxs->memory;
+    benum->grayscale = false;
+    if (is_jpeg) {
+        int code = px_jpeg_init(benum, params, par);
+        if (code < 0 || code == pxNeedData)
+            return_error(code);
+        depth = params->depth;
+        num_components = (params->color_space == eGray ? 1 : 3);
+        if (num_components == 3 && pxgs->color_space == eGray)
+            benum->grayscale = true;
+    } else {
+        px_gstate_t *pxgs = pxs->pxgs;
+        depth = "\001\004\010"[bar->depth];
+        num_components = (pxgs->color_space == eGray ? 1 : 3);
+        params->width = bar->width;
+        params->height = bar->height;
+        params->depth = depth;
+        params->color_space = pxgs->color_space;
+    }
 
-    params->width = par->pv[2]->value.i;
-    params->height = par->pv[3]->value.i;
-    params->depth = depth;
-    params->color_space = pxgs->color_space;
-    if (par->pv[0]->value.i == eIndexedPixel) {
+    if (bar->mapping == eIndexedPixel && !is_jpeg) {
         if (pxgs->palette.data == 0)
             return_error(errorMissingPalette);
         if (pxgs->palette.size != (1 << depth) * num_components)
@@ -160,25 +273,45 @@ begin_bitmap(px_bitmap_params_t * params, px_bitmap_enum_t * benum,
         num_components = 1;
     } else
         params->indexed = false;
-    params->dest_width = real_value(par->pv[4], 0);
-    params->dest_height = real_value(par->pv[4], 1);
+
+    params->dest_width = bar->dest_width;
+    params->dest_height = bar->dest_height;
     benum->data_per_row =
-        round_up(params->width * params->depth * num_components, 8) >> 3;
-    benum->mem = pxs->memory;
-    benum->initialized = false;
+        ((params->width * params->depth * num_components) + 7) >> 3;
     return 0;
 }
 
+/* Extract the parameters for reading a bitmap image or raster pattern. */
+/* Attributes: pxaColorMapping, pxaColorDepth, pxaSourceWidth, */
+/* pxaSourceHeight, pxaDestinationSize. */
 static int
-stream_error(stream_state * st, const char *str)
+begin_rebuffered_bitmap(px_bitmap_params_t * params, px_bitmap_enum_t * benum,
+                        const px_bitmap_args_t * bar, const px_state_t * pxs, bool is_jpeg, px_args_t * par)
 {
-    dmprintf1(st->memory, "pxl stream error %s\n", str);
-    return 0;
+    int code = begin_bitmap(params, benum, bar, pxs, is_jpeg, par);
+
+    if (code < 0 || code == pxNeedData)
+        return_error(code);
+    
+    if (params->width == 1 && params->height > 1) {
+        benum->rebuffered_data_per_row = benum->data_per_row * params->height;
+        benum->rebuffered = 1;
+        benum->rebuffered_width = params->height;
+        benum->rebuffered_height = params->width;
+        benum->rebuffered_row_pos = 0;
+    } else {
+        benum->rebuffered_data_per_row = benum->data_per_row;
+        benum->rebuffered = 0;
+        benum->rebuffered_width = params->width;
+        benum->rebuffered_height = params->height;
+    }
+
+    return code;
 }
 
 static int
 read_jpeg_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata,
-                      px_args_t * par)
+                      px_args_t * par, bool last)
 {
     uint data_per_row = benum->data_per_row;    /* jpeg doesn't pad */
     uint avail = par->source.available;
@@ -199,6 +332,9 @@ read_jpeg_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata,
         return 0;
     }
 
+    if (last)
+        return_error(errorIllegalDataLength);
+
     if (!benum->initialized) {
         jpeg_decompress_data *jddp = &(benum->jdd);
 
@@ -212,6 +348,9 @@ read_jpeg_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata,
         jddp->memory = ss->jpeg_memory = benum->mem;
         /* set this early for safe error exit */
         jddp->scanline_buffer = NULL;
+        jddp->PassThrough = 0;
+        jddp->PassThroughfn = 0;
+        jddp->device = (void *)0;
         if (gs_jpeg_create_decompress(ss) < 0)
             return_error(errorInsufficientMemory);
         (*s_DCTD_template.init) ((stream_state *) ss);
@@ -254,7 +393,7 @@ read_jpeg_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata,
 
 static int
 read_uncompressed_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata,
-                              px_args_t * par)
+                              px_args_t * par, bool last)
 {
     int code;
     uint avail = par->source.available;
@@ -274,6 +413,9 @@ read_uncompressed_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata,
     if (par->source.position >= data_per_row_padded * par->pv[1]->value.i)
         return 0;
 
+    if (last)
+        return_error(errorIllegalDataLength);
+    
     if (avail >= data_per_row_padded && pos_in_row == 0) {
         /* Use the data directly from the input buffer. */
         *pdata = (byte *) data;
@@ -293,7 +435,7 @@ read_uncompressed_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata,
 }
 
 static int
-read_rle_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata, px_args_t * par)
+read_rle_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata, px_args_t * par, bool last)
 {
     stream_RLD_state *ss = (&benum->rld_stream_state);
     uint avail = par->source.available;
@@ -314,6 +456,9 @@ read_rle_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata, px_args_t * par)
     if (par->source.position >= data_per_row_padded * par->pv[1]->value.i)
         return 0;
 
+    if (last)
+        return_error(errorIllegalDataLength);
+    
     if (!benum->initialized) {
         ss->EndOfData = false;
         ss->templat = &s_RLD_template;
@@ -404,7 +549,7 @@ read_rle_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata, px_args_t * par)
  */
 static int
 read_deltarow_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata,
-                          px_args_t * par)
+                          px_args_t * par, bool last)
 {
     deltarow_state_t *deltarow = &benum->deltarow_state;
     uint avail = par->source.available;
@@ -418,6 +563,9 @@ read_deltarow_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata,
         return 0;
     }
 
+    if (last)
+        return_error(errorIllegalDataLength);
+
     /* initialize at begin of image */
     if (!benum->initialized) {
         /* zero seed row */
@@ -544,41 +692,115 @@ read_deltarow_bitmap_data(px_bitmap_enum_t * benum, byte ** pdata,
  * Attributes: pxaStartLine (ignored), pxaBlockHeight, pxaCompressMode.
  */
 static int
-read_bitmap(px_bitmap_enum_t * benum, byte ** pdata, px_args_t * par)
+read_bitmap(px_bitmap_enum_t * benum, byte ** pdata, px_args_t * par, bool last)
 {
-    benum->compress_type = par->pv[2]->value.i;
+    /* Changing compression within an image is unimplemented */
+    /* but for now we return an error. */
+    if ((benum->compress_type != par->pv[2]->value.i) && (par->pv[2]->value.i != eNoCompression))
+        return_error(errorIllegalAttributeValue);
+    else
+        benum->compress_type = par->pv[2]->value.i;
     switch (benum->compress_type) {
         case eRLECompression:
-            return read_rle_bitmap_data(benum, pdata, par);
+            return read_rle_bitmap_data(benum, pdata, par, last);
         case eJPEGCompression:
-            return read_jpeg_bitmap_data(benum, pdata, par);
+            return read_jpeg_bitmap_data(benum, pdata, par, last);
         case eDeltaRowCompression:
-            return read_deltarow_bitmap_data(benum, pdata, par);
+            return read_deltarow_bitmap_data(benum, pdata, par, last);
         case eNoCompression:
-            return read_uncompressed_bitmap_data(benum, pdata, par);
+            return read_uncompressed_bitmap_data(benum, pdata, par, last);
         default:
             break;
     }
     return -1;
 }
 
+static int read_rebuffered_bitmap(px_bitmap_enum_t * benum, byte ** pdata, px_args_t * par)
+{
+    int code;
+
+    if (!benum->rebuffered)
+        return read_bitmap(benum, pdata, par, /* last */ false);
+
+    {
+        int w = benum->rebuffered_width;
+        byte *rowptr = *pdata;
+        byte *data2 = rowptr + benum->rebuffered_row_pos * benum->data_per_row;
+        for (; benum->rebuffered_row_pos < w; benum->rebuffered_row_pos++) {
+            byte *data3 = data2;
+            code = read_bitmap(benum, &data3, par, /* last */ false);
+            if (code == 0) {
+                return 0;
+            } else if (code == 1) {
+                /* got a scanline! (well, 1 pixel really, cos we are flipping) */
+            } else {
+                if (par->source.available == 0)
+                    return pxNeedData;
+                return code;
+            }
+            if (data3 != data2) {
+                memcpy(data2, data3, benum->data_per_row);
+            }
+            data2 += benum->data_per_row;
+        }
+        benum->rebuffered_row_pos = 0;
+    }
+    return code;
+}
+
+
 /* ---------------- Image operators ---------------- */
 
 const byte apxBeginImage[] = {
     pxaColorMapping, pxaColorDepth, pxaSourceWidth, pxaSourceHeight,
     pxaDestinationSize, 0, 0
 };
+
 int
 pxBeginImage(px_args_t * par, px_state_t * pxs)
+{
+    px_image_enum_t *pxenum;
+    px_bitmap_args_t bi_args;
+
+    pxenum =
+        gs_alloc_struct(pxs->memory, px_image_enum_t,
+                        &st_px_image_enum, "setup_bitmap(pxenum)");
+
+    if (pxenum == 0)
+        return_error(errorInsufficientMemory);
+
+    /*
+     * store a copy of the args.  For JPEG these parameters may be
+     * revised depending on the color parameters in the JPEG data.
+     * We defer image set up until we read the image data.
+     */
+    
+    bi_args.mapping     = par->pv[0]->value.i;
+    bi_args.depth       = par->pv[1]->value.i;
+    bi_args.width       = par->pv[2]->value.i;
+    bi_args.height      = par->pv[3]->value.i;
+    bi_args.dest_width  = real_value(par->pv[4], 0);
+    bi_args.dest_height = real_value(par->pv[4], 1);
+
+    pxenum->bi_args = bi_args;
+    pxenum->enum_started = false;
+    pxenum->benum.initialized = false;
+    pxs->image_enum = pxenum;
+    
+    return 0;
+}
+
+static int
+px_begin_image(px_state_t * pxs, bool is_jpeg, px_args_t * par)
 {
     gs_point origin;
     px_bitmap_params_t params;
-    px_bitmap_enum_t benum;
-    gs_state *pgs = pxs->pgs;
+    gs_gstate *pgs = pxs->pgs;
     px_gstate_t *pxgs = pxs->pxgs;
-    px_image_enum_t *pxenum;
+    px_image_enum_t *pxenum = pxs->image_enum;
+    px_bitmap_enum_t *pbenum = &pxenum->benum;
     int code;
-
+    
     if (gs_currentpoint(pgs, &origin) < 0)
         return_error(errorCurrentCursorUndefined);
     /*
@@ -600,34 +822,28 @@ pxBeginImage(px_args_t * par, px_state_t * pxs)
     code = px_set_halftone(pxs);
     if (code < 0)
         return code;
-    code = begin_bitmap(&params, &benum, par, pxs);
-    if (code < 0)
-        return code;
-    pxenum =
-        gs_alloc_struct(pxs->memory, px_image_enum_t,
-                        &st_px_image_enum, "setup_bitmap(pxenum)");
-
-    if (pxenum == 0)
-        return_error(errorInsufficientMemory);
-    {
-        pxenum->raster = round_up(benum.data_per_row, align_bitmap_mod);
-        pxenum->row = gs_alloc_byte_array(pxs->memory, 1, pxenum->raster,
-                                          "pxReadImage(row)");
-        if (pxenum->row == 0)
-            code = gs_note_error(errorInsufficientMemory);
-        else
-            code =
-                px_image_color_space(&pxenum->image, &params,
-                                     (const gs_string *)&pxgs->palette, pgs);
-    }
+    code = begin_rebuffered_bitmap(&params, pbenum, &pxenum->bi_args, pxs, is_jpeg, par);
+    if (code < 0 || code == pxNeedData)
+        return_error(code);
+    
+    pxenum->row = gs_alloc_byte_array(pxs->memory, 1, pbenum->rebuffered_data_per_row,
+                                      "pxReadImage(row)");
+    if (pxenum->row == 0)
+        code = gs_note_error(errorInsufficientMemory);
+    else
+        code =
+            px_image_color_space(&pxenum->image, &params,
+                                 (const gs_string *)&pxgs->palette, pgs);
+    
     if (code < 0) {
         gs_free_object(pxs->memory, pxenum->row, "pxReadImage(row)");
         gs_free_object(pxs->memory, pxenum, "pxReadImage(pxenum)");
         return code;
     }
+    
     /* Set up the image parameters. */
-    pxenum->image.Width = params.width;
-    pxenum->image.Height = params.height;
+    pxenum->image.Width = pbenum->rebuffered_width;
+    pxenum->image.Height = pbenum->rebuffered_height;
     {
         gs_matrix imat, dmat;
 
@@ -635,6 +851,10 @@ pxBeginImage(px_args_t * par, px_state_t * pxs)
         /* We also need to account for the upside-down H-P */
         /* coordinate system. */
         gs_make_scaling(params.width, params.height, &imat);
+        if (pbenum->rebuffered) {
+            imat.xy = imat.xx; imat.xx = 0;
+            imat.yx = imat.yy; imat.yy = 0;
+        }
         gs_make_translation(origin.x, origin.y, &dmat);
         gs_matrix_scale(&dmat, params.dest_width, params.dest_height, &dmat);
         /* The ImageMatrix is dmat' * imat. */
@@ -646,13 +866,16 @@ pxBeginImage(px_args_t * par, px_state_t * pxs)
 
     code = pl_begin_image(pgs, &pxenum->image, &pxenum->info);
     if (code < 0) {
-        gs_free_object(pxs->memory, pxenum->row, "pxReadImage(row)");
-        gs_free_object(pxs->memory, pxenum, "pxBeginImage(pxenum)");
+        /* This procedure will be re-invoked if we are remapping the
+           color so don't free the resources */
+        if (code != gs_error_Remap_Color) {
+            gs_free_object(pxs->memory, pxenum->row, "pxReadImage(row)");
+            gs_free_object(pxs->memory, pxenum, "pxBeginImage(pxenum)");
+        }
         return code;
     }
-    pxenum->benum = benum;
-    pxs->image_enum = pxenum;
     return 0;
+
 }
 
 const byte apxReadImage[] = {
@@ -664,22 +887,45 @@ int
 pxReadImage(px_args_t * par, px_state_t * pxs)
 {
     px_image_enum_t *pxenum = pxs->image_enum;
-
+    int code = 0;
+    
     if (par->pv[1]->value.i == 0)
         return 0;               /* no data */
     /* Make a quick check for the first call, when no data is available. */
     if (par->source.available == 0)
         return pxNeedData;
+    if (!pxenum->enum_started) {
+        bool is_jpeg = par->pv[2]->value.i == eJPEGCompression;
+        pxenum->benum.compress_type = par->pv[2]->value.i;
+        code = px_begin_image(pxs, is_jpeg, par);
+        if (code < 0 || code == pxNeedData)
+            return code;
+        pxenum->enum_started = true;
+    }
     for (;;) {
         byte *data = pxenum->row;
-        int code = read_bitmap(&pxenum->benum, &data, par);
-
+        int code = read_rebuffered_bitmap(&pxenum->benum, &data, par);
         if (code != 1)
             return code;
+
+        if (pxenum->benum.grayscale) {
+            int i;
+            for (i = 2; i < pxenum->benum.rebuffered_data_per_row; i+=3) {
+                int r = data[i-2];
+                int g = data[i-1];
+                int b = data[i];
+                /* we think this simple conversion is commensurate
+                   with typical XL business graphics use cases. */
+                int gray = (r + g + b) / 3;
+                data[i-2] = data[i-1] = data[i] = gray;
+            }
+        }
+
         code = pl_image_data(pxs->pgs, pxenum->info, (const byte **)&data, 0,
-                             pxenum->benum.data_per_row, 1);
+                             pxenum->benum.rebuffered_data_per_row, 1);
         if (code < 0)
             return code;
+        
         pxs->have_page = true;
     }
 }
@@ -717,6 +963,7 @@ struct px_pattern_enum_s
     int32_t pattern_id;
     pxePatternPersistence_t persistence;
     px_pattern_t *pattern;
+    int lines_rendered;
 };
 
 gs_private_st_simple(st_px_pattern_enum, px_pattern_enum_t,
@@ -734,14 +981,25 @@ pxBeginRastPattern(px_args_t * par, px_state_t * pxs)
     px_pattern_t *pattern;
     px_pattern_enum_t *pxenum;
     px_bitmap_enum_t benum;
+    px_bitmap_args_t bi_args;
     byte *data;
     uint psize;
     byte *pdata;
-
-    int code = begin_bitmap(&params, &benum, par, pxs);
+    int code;
     static const gs_memory_struct_type_t st_px_pattern =
         { sizeof(px_pattern_t), "", 0, 0, 0, 0, 0 };
 
+    
+    bi_args.mapping     = par->pv[0]->value.i;
+    bi_args.depth       = par->pv[1]->value.i;
+    bi_args.width       = par->pv[2]->value.i;
+    bi_args.height      = par->pv[3]->value.i;
+    bi_args.dest_width  = real_value(par->pv[4], 0);
+    bi_args.dest_height = real_value(par->pv[4], 1);
+    
+
+    code = begin_bitmap(&params, &benum, &bi_args, pxs, false, par);
+
     if (code < 0)
         return code;
     rc_alloc_struct_1(pattern, px_pattern_t, &st_px_pattern, mem,
@@ -769,8 +1027,10 @@ pxBeginRastPattern(px_args_t * par, px_state_t * pxs)
         return_error(errorInsufficientMemory);
     }
     pxenum->benum = benum;
+    pxenum->benum.initialized = false;
     pxenum->pattern_id = par->pv[5]->value.i;
     pxenum->persistence = par->pv[6]->value.i;
+    pxenum->lines_rendered = 0;
     pattern->params = params;
     pattern->palette.data = pdata;
     pattern->palette.size = psize;
@@ -791,33 +1051,42 @@ pxReadRastPattern(px_args_t * par, px_state_t * pxs)
 {
     px_pattern_enum_t *pxenum = pxs->pattern_enum;
     int code;
-    uint input_per_row = round_up(pxenum->benum.data_per_row, 4);
+    byte *plimit = pxenum->pattern->data +
+        (pxenum->benum.data_per_row * pxenum->pattern->params.height);
+
+    if (par->pv[1]->value.i == 0)
+        return 0;               /* no data */
+
+    /* first call */
+    if (par->source.available == 0)
+        pxenum->lines_rendered = 0;
 
-    if (par->pv[3])
-        input_per_row =
-            round_up(pxenum->benum.data_per_row, par->pv[3]->value.i);
-    /* Make a quick check for the first call, when no data is available. */
-    if (par->source.available == 0 && par->pv[1]->value.i != 0)
-        return pxNeedData;
-    /* emulate hp bug */
     {
         pxeCompressMode_t c = par->pv[2]->value.i;
 
-        if (c == eDeltaRowCompression || c == eJPEGCompression)
-            input_per_row = pxenum->benum.data_per_row;
+        if (!pxenum->benum.initialized)
+            pxenum->benum.compress_type = c;
+
     }
+    
     for (;;) {
         byte *data = pxenum->pattern->data +
-            (par->pv[0]->value.i + par->source.position / input_per_row)
-            * pxenum->benum.data_per_row;
+            (par->pv[0]->value.i + pxenum->lines_rendered) * pxenum->benum.data_per_row;
         byte *rdata = data;
 
-        code = read_bitmap(&pxenum->benum, &rdata, par);
+        if (data > plimit)
+            return_error(gs_error_rangecheck);
+
+        code = read_bitmap(&pxenum->benum, &rdata, par, /* last */ data == plimit);
         if (code != 1)
             break;
+
+        pxenum->lines_rendered++;
+
         if (rdata != data)
             memcpy(data, rdata, pxenum->benum.data_per_row);
     }
+
     return code;
 }
 
@@ -890,7 +1159,7 @@ pxScanLineRel(px_args_t * par, px_state_t * pxs)
                                  * We use the separate variable source.count to keep track of
                                  * the number of x-pairs left in the scan line.
                                  */
-    gs_state *pgs = pxs->pgs;
+    gs_gstate *pgs = pxs->pgs;
     bool big_endian = pxs->data_source_big_endian;
     const byte *data = par->source.data;
     pxeDataType_t
